(************************************************************************)
(*                                                                      *)
(*                    Version 2.1a without error checks                 *)
(*                                                                      *)
(************************************************************************)


IMPLEMENTATION MODULE SETs;

(************************************************************************)
(*                                                                      *)
(*                              S E T s                                 *)
(*                                                                      *)
(************************************************************************)
(*                                                                      *)
(*           Version 2.1a for Sun/4, SunOS 4.1.1, m2c-Compiler          *)
(*     Version 2.1b for IBM-PC/AT, MSDOS 3.30, Taylor Modula-2 3.00     *)
(*     Version 2.1c for IBM-PC/AT, MSDOS 3.30, Fitted Modula-2 2.00     *)
(*                                                                      *)
(*                  (c) Richard Breuer [BR] 1990 - 1992                 *)
(*            e-mail: ricki@rwthi3.informatik.rwth-aachen.de            *)
(*              mail: Brunssumstr. 6, 5100 Aachen, Germany              *)
(*                                                                      *)
(*          based on the module LLSetRel (c) Thomas Brandes 1984        *)
(*                                                                      *)
(************************************************************************)
(*                                                                      *)
(* Notes:                                                               *)
(*                                                                      *)
(*   The generic module  SETs realizes sets in their mathematical sense *)
(*   as objects which may contain several  elements each one at a time. *)
(*   SETs exports the usual operations on sets.                         *)
(*                                                                      *)
(*   Genericity is implemented by using the generic Modula-2 data types *)
(*   ARRAY OF BYTE  and  ADDRESS.  There is only  one instance of  SETs *)
(*   during  the runtime of an  application no matter how many sets  of	*)
(*   various kinds will be defined. The major disadvantage of the usage *)
(*   of the generic  data types is  that the  compiler has no chance to	*)
(*   check for data  type consistency at compile time.  So the applica-	*)
(*   tion must be careful not to affect unwanted side effects.          *)
(*                                                                      *)
(*   Note that the  deletion of elements is  implemeted just by marking *)
(*   elements as deleted. Their allocated memory space will be released *)
(*   when SETCompact is called.                                         *)
(*                                                                      *)
(*   Since  version 2.0  SETs  performs  the compacting  itself so that	*)
(*   there is no need for any application to call SETCompact anymore.	*)
(*                                                                      *)
(************************************************************************)
(*                                                                      *)
(* Revisions:                                                         	*)
(*                                                                      *)
(*   01/27/92	Version 2.1						*)
(*   01/10/92	SETKill: All error checks are made in SETClear [BR]	*)
(*   01/10/92   A new local procedure Get allows the implementation of	*)
(*		SETGet without searching in the tree [BR]		*)
(*   01/10/92	Insert: Check the filling criteria before inserting the	*)
(*		element. This must be done after the change of 01/03/92	*)
(*		in SETExtractAnyElement in order to make Insert compact	*)
(*		a set after many ExtractAny's [BR]			*)
(*   01/03/91	SETGet: Check for Set^.N=0 before traversing the set	*)
(*		[BR]							*)
(*   01/03/92	SETExtractAnyElement does not compact anymore [BR]	*)
(*   01/03/92	Reduction of the number of compactifications by in-	*)
(*		creasing the constants [BR]				*)
(*   11/12/91	SETRangeLoop: Correction of an error in the algorithm	*)
(*		for finding the smallest SETEqual-Element [BR]		*)
(*   11/01/91	Version 2.0 [BR]					*)
(*   11/01/91	Automatic compression of the set tree if one of two	*)
(*		criteria is true:					*)
(*		1. if the fill level of the set falls under the value	*)
(*		   FillingLimit percent					*)
(*		2. if the depth of the tree differs more than Balance-	*)
(*		   limit nodes from the optimal depth given by the	*)
(*		   value of trunc(1+ld n)				*)
(*		These two criteria will only be checked if the set has	*)
(*		more than FillingCount elements. Otherwise the compac-	*)
(*		ting is too expensive [BR]				*)
(*   10/31/91	SETExtractAnyElement: Check for Set^.N=0 before traver-	*)
(*		sing the set [BR]					*)
(*   09/06/91	SETExtractAnyElement: Call SETCheckConsistency in the	*)
(*		CHK version if it is clear that that the set is not in	*)
(*		a consistent state [BR]					*)
(*   08/29/91	Code simplification in DeleteElement and SETIn [BR]	*)
(*   08/29/91   see DEFINITION MODULE [BR]                              *)
(*   08/14/91	Corrections of some comments [BR]			*)
(*   08/03/91	Corrections of some comments [BR]			*)
(*   06/10/91   SETCopy: the source set is no longer compacted before   *)
(*              copying because the elements can be found in a sorted   *)
(*              way in the linear list [BR]                             *)
(*   05/30/91   Corrected a wrong procedure name in SETCompact [BR]     *)
(*   05/29/91   Corrected a wrong procedure name in SETIsDisjunct [BR]  *)
(*   05/29/91   Optimized SETIsSubset by adding a cardinality check [BR]*)
(*   05/29/91   see DEFINITION MODULE [BR]                              *)
(*   05/29/91   Corrected a wrong procedure name in SETUnion [BR]       *)
(*   05/14/91   Added the CreateFor/InsertUptoN checks [BR]             *)
(*   05/12/91   Version 1.1c	                                       	*)
(*   05/02/91   SETExtractAnyElement: now call DeleteElement instead of *)
(*              SETRemoveElement (effiziency+) [BR]                     *)
(*   05/02/91   Removed local procedure RemoveElemen; call DeleteElement*)
(*              instead [BR]                                            *)
(*   02/27/91   Versions 1.1a,1.1b                                 	*)
(*   02/27/91   Removed unused declarations [BR]                        *)
(*   02/27/91   SETCreate: presetting of RangeCmp inserted [BR]         *)
(*   02/27/91   SETExtractAnyElement: Error because SETGet was called   *)
(*              without setting RangeCmp before [BR]                    *)
(*   02/25/91   see DEFINTION MODULE [BR]                               *)
(*   02/24/91   Inserted error checks at allocates [BR]                 *)
(*   02/23/91   Error detected: ClassN was not initialized [BR]         *)
(*   02/23/91   SETCompact: the call to SETSort can be left out because *)
(*              the sorting of the elements is insured by SETGet [BR]   *)
(*   02/21/91   SETGotoElement: Error, because Found=TRUE was also set  *)
(*              if an element had been deleted [BR]                     *)
(*   02/20/91   The declaration of SetType must be enlarged to contain  *)
(*              the RangeCmp value in order to realize the SETRangeLoop *)
(*              procedure. RangeCmp will always contain the LoopCmpProc *)
(*              in a SETRangeLoop/SETGet loop [BR]                      *)
(*   02/20/91   FindData: parameter Result instead of Found minimizes   *)
(*              the calls to the Compare procedure [BR]                 *)
(*   02/20/91   Changed the overall behaviour: the elements will from	*)
(*		now on be linked in a sorted order in the linear list.	*)
(*		Thus SETGet will return the elements sorted [BR]	*)
(*   02/07/91   SETClear: Error found: deallocation of just TSIZE(Set-  *)
(*              Type) bytes instead of TSIZE(ELementType) bytes for     *)
(*              TreeHead [BR]                                           *)
(*   02/07/91   see DEFINITION MODULE [BR]                              *)
(*   12/11/90   SETExtractAnyElement: Optimization by not calling SET-  *)
(*              Loop in most cases [BR]                                 *)
(*   09/20/90   Realization of the interface with the data type ARRAY   *)
(*              OF BYTE insted of ADDRESS. Rewinding the modification   *)
(*              of 09/13/90 [BR]                                        *)
(*   09/17/90   More error checks [BR]                                  *)
(*   09/13/90   Using compiler switches instead of the CCALL call [BR]  *)
(*   09/13/90   see DEFINITION MODULE [BR]                              *)
(*   09/12/90   Simplication of the interface by demanding that all     *)
(*              set elements have the same byte length [BR]             *)
(*   09/12/90   Parameters of compare procedures now have ADDRESS para- *)
(*              meters instead of ARRAY OF BYTE [BR]                    *)
(*   09/11/90   Comments and some corrections [BR]                      *)
(*   09/09/90   Versions 1.0a,1.0b                                	*)
(*                                                                      *)
(************************************************************************)

  FROM SYSTEM IMPORT
    ADDRESS,
    CCALL,
    ADR,
    BYTE,
    TSIZE;



  FROM Storage IMPORT
    ALLOCATE,
    DEALLOCATE;


  CONST
    MaxClasses = 256;
                                (* maximal count of set classes which   *)
                                (* may be defined at runtime            *)

    ModuleName = "SETs 2.1a [no checks]";

    FillingCount = 31;
				(* the automatic compression will only	*)
				(* be invoked if there are more than	*)
				(* FillingCount elements in the set	*)
				
    FillingLimit = 50;
				(* Compaction takes place if the rela-	*)
				(* tion of deleted elements to not-dele-*)
				(* ted ones is worse than this value	*)

    BalanceLimit = 10;
				(* Compaction takes place if there is	*)
				(* a leaf in the tree whose height	*)
				(* differs from the optimal height more	*)
				(* than BalanceHeight nodes		*)

  TYPE
    ElementTypePtr = POINTER TO ElementType;
                                (* pointer to one set element           *)

    ElementType = RECORD
                                (* this record contains the data stored *)
                                (* with each element                    *)
      DataPtr: ADDRESS;
                                (* - pointer to the data bytes          *)
      Deleted: BOOLEAN;
                                (* - TRUE if the element is marked as   *)
                                (*   deleted                            *)
      Left   : ElementTypePtr;
                                (* - pointer to left part search tree   *)
      Right  : ElementTypePtr;
                                (* - pointer to right part search tree  *)
      Thread : ElementTypePtr
                                (* - successor in the linear list       *)
    END;

    SETType = POINTER TO SetType;
                                (* generic set type                     *)

    SetType = RECORD
                                (* this record contains the data stored *)
                                (* with each set                        *)

      TreeHead  : ElementTypePtr;
                                (* - pointer to the root of the binary  *)
                                (*   search tree                        *)
      ListThread: ElementTypePtr;
                                (* - pointer to the start of the linear *)
                                (*   list which contains all elements   *)
                                (*   in an ordered way. Therefore List- *)
                                (*   Thread will always point to the    *)
                                (*   smallest element                   *)
      Current   : ElementTypePtr;
                                (* - pointer to the current element     *)
                                (*   when a loop is in progress         *)
      RangeCmp  : SETCompareProcType;
                                (* - temporary storage for the compare  *)
                                (*   procedure when SETRangeLoop/SETGet *)
                                (*   loop is in progress                *)
      ClassID   : CARDINAL;
                                (* - number of the set class            *)
      DeletedN  : CARDINAL;
                                (* - count for the elements which are   *)
                                (*   marked as deleted                  *)
      N         : CARDINAL;
                                (* - count for (logical) set elements   *)
      Depth     : CARDINAL;
                                (* - current depth of the set tree	*)
    END;

    ClassType = RECORD
                                (* this record contains all information *)
                                (* stored for one class                 *)
      Cmp   : SETCompareProcType;
                                (* - compare procedure                  *)
      Bytes: CARDINAL
                                (* - length of an element in bytes      *)
    END;

  VAR
    Classes: ARRAY [0..MaxClasses-1] OF ClassType;
                                (* this array contains the information  *)
                                (* for the set classes                  *)

    ClassN: CARDINAL;
                                (* number of currently defined set      *)
                                (* classes                              *)

  (**********************************************************************)
  (*                                                                    *)
  (*                          Local Procedures                          *)
  (*                                                                    *)
  (**********************************************************************)


  PROCEDURE Copy(    From: ADDRESS;
                     To  : ADDRESS;
                     Len : CARDINAL);
  (* copies Len bytes From -> To. There seems to be no clean Modula-2
     solution of this problem *)


  BEGIN
    (* call the C procedure 'bcopy' *)
    CCALL("bcopy",From,To,Len)
  END Copy;

  PROCEDURE OptimalDepth(    N: CARDINAL): CARDINAL;
  
  BEGIN
    (* This procedures uses an if cascade instead of a REAL calculation for
       efficiency reasons (at least I hope that this is more efficient
       than something like "trunc(1.0+log2(float(n)))" *)
    IF (N=0)     THEN RETURN  0 END;
    IF (N<2)     THEN RETURN  1 END;
    IF (N<4) 	 THEN RETURN  2 END;
    IF (N<8) 	 THEN RETURN  3 END;
    IF (N<16) 	 THEN RETURN  4 END;
    IF (N<32) 	 THEN RETURN  5 END;
    IF (N<64) 	 THEN RETURN  6 END;
    IF (N<128) 	 THEN RETURN  7 END;
    IF (N<256)   THEN RETURN  8 END;
    IF (N<512)   THEN RETURN  9 END;
    IF (N<1024)  THEN RETURN 10 END;
    IF (N<2048)  THEN RETURN 11 END;
    IF (N<4096)  THEN RETURN 12 END;
    IF (N<8192)  THEN RETURN 13 END;
    IF (N<16384) THEN RETURN 14 END;
    IF (N<32768) THEN RETURN 15 END;
    IF (N<65536)      THEN RETURN 16 END;
    IF (N<131072)     THEN RETURN 17 END;
    IF (N<262144)     THEN RETURN 18 END;
    IF (N<524288)     THEN RETURN 19 END;
    IF (N<1048576)    THEN RETURN 20 END;
    IF (N<2097152)    THEN RETURN 21 END;
    IF (N<4194304)    THEN RETURN 22 END;
    IF (N<8388608)    THEN RETURN 23 END;
    IF (N<16777216)   THEN RETURN 24 END;
    IF (N<33554432)   THEN RETURN 25 END;
    IF (N<67108864)   THEN RETURN 26 END;
    IF (N<134217728)  THEN RETURN 27 END;
    IF (N<268435456)  THEN RETURN 28 END;
    IF (N<536870912)  THEN RETURN 29 END;
    IF (N<1073741824) THEN RETURN 30 END;
    IF (N<2147483648) THEN RETURN 31 END;
    RETURN(32);
  END OptimalDepth;

  PROCEDURE MakeCell(VAR N    : ElementTypePtr;
                         Data : ADDRESS;
                         Bytes: CARDINAL);
  (* creates a new set element node and copies the data bytes *)

  BEGIN
    (* allocate space for the new node and its data *)
    ALLOCATE(N,TSIZE(ElementType));
    ALLOCATE(N^.DataPtr,Bytes);
    (* copy the data bytes *)
    Copy(Data,N^.DataPtr,Bytes);
    N^.Deleted := FALSE;
    N^.Left := NIL;
    N^.Right := NIL;
    N^.Thread := NIL
  END MakeCell;

  PROCEDURE FindData(VAR Root   : ElementTypePtr;
                         Data   : ADDRESS;
                         Compare: SETCompareProcType;
                     VAR Prev   : ElementTypePtr;
	 	     VAR Depth  : CARDINAL;
                     VAR Result : SETCompareResultType);
  (* Find Data in the (partial) binary search tree under Root. Root returns
     a pointer either to the element which contains Data (if Result=Equal)
     or the node under which a new element has to be inserted (if Result#
     Equal). Compare is the compare procedure that will be used. Prev will
     return a pointer to the element that is the direct predecessor of Root
     in the sorting order. It therefore holds the constraint Prev^.Thread
     =Root. If Result#Equal, Result will return how Data has to be inserted
     under Root. Depth returns the depth of the element which has been
     found. If the tree is empty, Depth=0 is returned. If the element
     could not be found, Depth contains the depth of the element under
     which a new element sould be inserted *)

  VAR
    Next: ElementTypePtr;

  BEGIN
    (* start with the specified root *)
    Next := Root;
    Prev := NIL;
    Depth := 0;
    (* loop until leaf is reached or data is found *)
    Result := SETLess;
    WHILE (Next<>NIL) AND (Result#SETEqual) DO
      INC(Depth);
      Root := Next;
      (* comparison *)
      Result := Compare(Data,Next^.DataPtr);
      CASE Result OF
        SETEqual  : ;
      | SETLess   : Next := Next^.Left;
        (* Prev stores the address of the node where the algorithm stepped
           down to the right for the last time. This element will at the end
           be the direct predecessor of the target node in the linear sorted
           list [algorithm by Andy Schuerr] *)
      | SETGreater: Prev := Next; Next := Next^.Right
      END
    END
  END FindData;

  PROCEDURE Insert(VAR Set : SETType;
                       Data: ADDRESS);
  (* insert Data into Set. No action if Data is already contained in set *)

  VAR
    Root,Prev: ElementTypePtr;
    New      : ElementTypePtr;
    Result   : SETCompareResultType;
    Depth    : CARDINAL;
    OptDepth : CARDINAL;

  BEGIN
    (* Check if the set must be compacted before an element is inserted.
       This check is done before the insertion to avoid long searches in
       empty sets *)
    IF (Set^.N+Set^.DeletedN>FillingCount) THEN
      IF (Set^.N>0) THEN
        IF (TRUNC(100.0*FLOAT(Set^.DeletedN)/FLOAT(Set^.N))>FillingLimit)
        THEN SETCompact(Set) END
      ELSE
        SETClear(Set)
      END
    END;
    (* Try to locate Data in Set *)
    Root := Set^.TreeHead;
    FindData(Root,Data,Classes[Set^.ClassID].Cmp,Prev,Depth,Result);
    IF (Result=SETEqual) THEN
      (* Remove deletion mark, if the element was found *)
      IF (Root^.Deleted) THEN
        DEC(Set^.DeletedN);
        INC(Set^.N);
        Root^.Deleted := FALSE
      END
    ELSE
      INC(Set^.N);
      (* allocate new space for Data *)
      MakeCell(New,Data,Classes[Set^.ClassID].Bytes);
      (* insert Data in the binary tree *)
      IF (Root=NIL) THEN
        (* preset header values if the element is the first in the set *)
        Set^.TreeHead := New;
        Set^.ListThread := New;
	Set^.Depth := 1
      ELSE
        IF (Result=SETLess) THEN
          Root^.Left := New;
          (* insert the new element before Root in the linear list because
             it is the direct predecessor *)
          IF (Prev=NIL) THEN
            Set^.ListThread := New ELSE Prev^.Thread := New
          END;
          New^.Thread := Root
        ELSE
          Root^.Right := New;
          (* insert the new element after Root in the linear list because
             it is the direct successor *)
          New^.Thread := Root^.Thread;
          Root^.Thread := New
	END;
	(* Check whether Set^.Depth must be updated or the set must be
	   compacted *)
	IF (Depth=Set^.Depth) THEN
	  (* Are there enough elements? *)
          IF (Set^.N+Set^.DeletedN>FillingCount) THEN
	    (* Calculate optimal depth *)
	    OptDepth := OptimalDepth(Set^.N+Set^.DeletedN);
	    (* Check the difference of the current and the optimal depth *)
	    IF (Depth+1-OptDepth>BalanceLimit) THEN SETCompact(Set)
	    				       ELSE INC(Set^.Depth) END
	  ELSE
	    INC(Set^.Depth)
	  END
	END
      END
    END
  END Insert;

  PROCEDURE InsertUptoN(VAR Set : SETType;
                            Data: ADDRESS);
  (* insert the next element in a CreateForN/InsertUptoN loop *)

  BEGIN
    (* store values at the current element *)
    WITH Set^.Current^ DO
      ALLOCATE(DataPtr,Classes[Set^.ClassID].Bytes);
      Copy(Data,DataPtr,Classes[Set^.ClassID].Bytes);
      Deleted := FALSE;
    END;
    (* goto successor *)
    Set^.Current := Set^.Current^.Thread
  END InsertUptoN;

  PROCEDURE DeleteElement(VAR Set  	 : SETType;
                              Data 	 : ADDRESS;
                          VAR Found	 : BOOLEAN);

  VAR
    Root,P: ElementTypePtr;
    D     : CARDINAL;
    Result: SETCompareResultType;

  BEGIN
    (* locate Data *)
    Root := Set^.TreeHead;
    FindData(Root,Data,Classes[Set^.ClassID].Cmp,P,D,Result);
    Found := (Result=SETEqual) AND (NOT Root^.Deleted);
    (* if found and not yet deleted: mark as deleted *)
    IF (Found) THEN
      DEC(Set^.N); INC(Set^.DeletedN); Root^.Deleted := TRUE;
      (* Check for reorganization *)
      IF (Set^.N+Set^.DeletedN>FillingCount) THEN
	(* Check the filling percentage *)
        IF (TRUNC(100.0*FLOAT(Set^.DeletedN)/FLOAT(Set^.N))>FillingLimit)
        THEN SETCompact(Set) END
      END
    END
  END DeleteElement;

  PROCEDURE In(    Set : SETType;
                   Data: ADDRESS): BOOLEAN;

  VAR
    Root,P: ElementTypePtr;
    D     : CARDINAL;
    Result: SETCompareResultType;

  BEGIN
    (* locate Data *)
    Root := Set^.TreeHead;
    FindData(Root,Data,Classes[Set^.ClassID].Cmp,P,D,Result);
    (* found if not marked as deleted *)
    RETURN (Result=SETEqual) AND (NOT Root^.Deleted)
  END In;

  PROCEDURE FullLoopCmpProc(    Data1,Data2: ADDRESS):
            SETCompareResultType;

  BEGIN RETURN(SETEqual) END FullLoopCmpProc;

  PROCEDURE Sort(VAR Set: SETType);
  (* create a sorted linear list from the binary search tree pointers *)

  VAR
    First: ElementTypePtr;

    PROCEDURE PartTreeSort(    Root : ElementTypePtr;
                               Succ : ElementTypePtr;
                           VAR First: ElementTypePtr);
    (* sort the partial tree under Root *)

    VAR
      Help: ElementTypePtr;

    BEGIN
      IF (Root^.Left=NIL) THEN
        First := Root
      ELSE
        PartTreeSort(Root^.Left,Root,First)
      END;
      IF (Root^.Right=NIL) THEN
        Root^.Thread := Succ
      ELSE
        PartTreeSort(Root^.Right,Succ,Help);
        Root^.Thread := Help
      END
    END PartTreeSort;

  BEGIN
    (* no action if list is empty (that means: set is empty) *)
    IF (Set^.ListThread#NIL) THEN
      (* sort the set; First will contain a pointer to the smallest set
         element after the call *)
      PartTreeSort(Set^.TreeHead,NIL,First);
      Set^.ListThread := First
    END
  END Sort;

  PROCEDURE Get(VAR Set  : SETType;
                VAR Ptr  : ElementTypePtr;
                VAR Found: BOOLEAN);
  (* Get computes a pointer to a next element in a loop *)

  VAR
    Last  : ElementTypePtr;
    Result: SETCompareResultType;

  BEGIN
    IF (Set^.N=0) THEN Found := FALSE; RETURN END;
    (* store current pointer *)
    Last := Set^.Current;
    Result := SETEqual;
    (* loop until there is a non-deleted element or until the end of the
       SETEqual range is reached *)
    WHILE (Set^.Current#NIL) AND (Result=SETEqual) DO
      IF (Set^.Current^.Deleted) THEN
        (* step to successor if deleted *)
        Set^.Current := Set^.Current^.Thread;
        (* check for still being in the SETEqual range *)
        IF (Set^.Current#NIL) THEN
          Result := Set^.RangeCmp(Set^.Current^.DataPtr,Last^.DataPtr)
        END
      ELSE
	(* The current pointer contains the next element *)
        Ptr := Set^.Current;
        Found := TRUE;
        (* go to sucessor *)
        Set^.Current := Set^.Current^.Thread;
        (* if there is a successor: check whether it is still in the
           SETEqual range. If not, the Loop/Get loop will terminate with
           the next call *)
        IF (Set^.Current#NIL) THEN
          IF (Set^.RangeCmp(Set^.Current^.DataPtr,Last^.DataPtr)#SETEqual) THEN
            Set^.Current := NIL
          END
        END;
        RETURN
      END
    END;
    Found := FALSE
  END Get;

  (**********************************************************************)
  (*                                                                    *)
  (*                        Exported Procedures                         *)
  (*                                                                    *)
  (**********************************************************************)

  PROCEDURE SETCreateClass(    CmpProc: SETCompareProcType;
                               Bytes  : CARDINAL;
                           VAR ClassID: CARDINAL);

  BEGIN
    (* store the values for the specified class *)
    Classes[ClassN].Cmp := CmpProc;
    Classes[ClassN].Bytes := Bytes;
    ClassID := ClassN;
    (* increment class counter *)
    INC(ClassN)
  END SETCreateClass;

  PROCEDURE SETCreate(VAR Set  : SETType;
                          Class: CARDINAL);

  BEGIN
    (* allocate some space for the set header *)
    ALLOCATE(Set,TSIZE(SetType));
    WITH Set^ DO
      DeletedN := 0;
      N := 0;
      Depth := 0;
      ListThread := NIL;
      TreeHead := NIL;
      Current := NIL;
      ClassID := Class;
      RangeCmp := FullLoopCmpProc
    END
  END SETCreate;

  PROCEDURE SETSetUndefined(VAR Set: SETType);

  BEGIN Set := NIL
  END SETSetUndefined;

  PROCEDURE SETIsDefined(    Set: SETType): BOOLEAN;

  BEGIN RETURN(Set#NIL)
  END SETIsDefined;

  PROCEDURE SETKill(VAR Set: SETType);

  BEGIN
    (* make set empty and deallocate its storage *)
    SETClear(Set);
    DEALLOCATE(Set,TSIZE(SetType));
    Set := NIL
  END SETKill;

  PROCEDURE SETClear(VAR Set: SETType);

  BEGIN
    WITH Set^ DO
      (* loop over all elements to deallocate their storage *)
      Current := ListThread;
      WHILE (Current#NIL) DO
        TreeHead := Current;
        Current := Current^.Thread;
        DEALLOCATE(TreeHead^.DataPtr,Classes[ClassID].Bytes);
        DEALLOCATE(TreeHead,TSIZE(ElementType))
      END;
      (* reinstall the starting status *)
      ListThread := NIL;
      TreeHead := NIL;
      Current := NIL;
      DeletedN := 0;
      N := 0
    END
  END SETClear;

  PROCEDURE SETInsert(VAR Set  : SETType;
                      VAR Data : ARRAY OF BYTE);

  BEGIN
    Insert(Set,ADR(Data))
  END SETInsert;

  PROCEDURE SETGetValue(    Set  : SETType;
		        VAR Data : ARRAY OF BYTE;
		        VAR Found: BOOLEAN);

  VAR
    Root  : ElementTypePtr;
    P     : ElementTypePtr;
    D     : CARDINAL;
    Result: SETCompareResultType;

  BEGIN
    (* find Data in the set. If it can be found, copy the contents of the
       set element onto the parameter *)
    Root := Set^.TreeHead;
    FindData(Root,ADR(Data),Classes[Set^.ClassID].Cmp,P,D,Result);
    Found := (Result=SETEqual) AND (NOT Root^.Deleted);
    IF (Found) THEN 
      Copy(Root^.DataPtr,ADR(Data),Classes[Set^.ClassID].Bytes) 
    END
  END SETGetValue;

  PROCEDURE SETReplaceValue(VAR Set  : SETType;
			    VAR Data : ARRAY OF BYTE;
			    VAR Found: BOOLEAN);

  VAR
    Root  : ElementTypePtr;
    P     : ElementTypePtr;
    D     : CARDINAL;
    Result: SETCompareResultType;

  BEGIN
    (* find Data in the set. If it can be found, copy the contents of the
       parameter onto the set element. The sorting of the set cannot be
       destroyed by this! *)
    Root := Set^.TreeHead;
    FindData(Root,ADR(Data),Classes[Set^.ClassID].Cmp,P,D,Result);
    Found := (Result=SETEqual) AND (NOT Root^.Deleted);
    IF (Found) THEN 
      Copy(ADR(Data),Root^.DataPtr,Classes[Set^.ClassID].Bytes) 
    END
  END SETReplaceValue;

  PROCEDURE SETIsEmpty(    Set: SETType): BOOLEAN;

  BEGIN
    RETURN(Set^.N=0)
  END SETIsEmpty;

  PROCEDURE SETCreateForN(VAR Set    : SETType;
                              ClassID: CARDINAL;
                              N      : CARDINAL);

  VAR
    Root: ElementTypePtr;

    PROCEDURE RecursiveCreate(VAR Root: ElementTypePtr;
                                  N   : CARDINAL);
    (* create a partial tree under root which can hold N elements *)

    BEGIN
      IF (N=0) THEN Root := NIL
      ELSE
        (* create node with dummy contents *)
        ALLOCATE(Root,TSIZE(ElementType));
        Root^.Deleted := FALSE;
        (* create subtrees *)
        RecursiveCreate(Root^.Left,(N DIV 2));
        RecursiveCreate(Root^.Right,((N-1) DIV 2))
      END
    END RecursiveCreate;

  BEGIN
    SETCreate(Set,ClassID);
    (* insert N dummy elements *)
    RecursiveCreate(Root,N);
    Set^.N := N;
    Set^.Depth := OptimalDepth(N);
    Set^.DeletedN := 0;
    Set^.ListThread := Root;
    Set^.TreeHead := Root;
    (* use Sort to create the linear list *)
    Sort(Set);
    (* make the first element the current one. SETInsertUptoN will start with
       replacing this element *)
    Set^.Current := Set^.ListThread
  END SETCreateForN;

  PROCEDURE SETInsertUptoN(VAR Set : SETType;
                           VAR Data: ARRAY OF BYTE);

  BEGIN
    InsertUptoN(Set,ADR(Data))
  END SETInsertUptoN;

  PROCEDURE SETRemoveElement(VAR Set : SETType;
                             VAR Data: ARRAY OF BYTE);

  VAR
    Found: BOOLEAN;

  BEGIN
    DeleteElement(Set,ADR(Data),Found)
  END SETRemoveElement;

  PROCEDURE SETDeleteElement(VAR Set  : SETType;
                             VAR Data : ARRAY OF BYTE;
                             VAR Found: BOOLEAN);

  BEGIN
    DeleteElement(Set,ADR(Data),Found)
  END SETDeleteElement;

  PROCEDURE SETExtractAnyElement(VAR Set  : SETType;
                                 VAR Data : ARRAY OF BYTE;
                                 VAR Found: BOOLEAN);

  VAR
    ElemPtr: ElementTypePtr;

  BEGIN
    IF (Set^.N=0) THEN Found := FALSE; RETURN END;
    (* first try to find a non-deleted element without starting at the
       beginning --> efficiency+ *)
    Set^.RangeCmp := FullLoopCmpProc;
    Get(Set,ElemPtr,Found);
    (* start again if there's no element *)
    IF (NOT Found) THEN
      Set^.Current := Set^.ListThread;
      (* find first non-deleted element *)
      Get(Set,ElemPtr,Found)
    END;
    (* Copy & Delete if found. Perfom no automatic compactification because
       this would decrease the efficiency of SETExtractAnyElement too much *)
    IF (Found) THEN
      Copy(ElemPtr^.DataPtr,ADR(Data),Classes[Set^.ClassID].Bytes);
      DEC(Set^.N); INC(Set^.DeletedN); ElemPtr^.Deleted := TRUE
    END
  END SETExtractAnyElement;

  PROCEDURE SETCompact(VAR Set: SETType);

  VAR
    HelpSet: SETType;
    Curr   : ElementTypePtr;
    R,R1   : ElementTypePtr;

  BEGIN
    (* create an optimal tree for N elements *)
    SETCreateForN(HelpSet,Set^.ClassID,Set^.N);
    (* loop over the source set *)
    R := Set^.ListThread;
    Curr := NIL;
    WHILE (R#NIL) DO
      (* insert a non-deleted element into the target set *)
      IF (NOT R^.Deleted) THEN
        (* If the current element is in R, its new value is stored in Curr.
           If R was a deleted element, Curr will point to NIL *)
        IF (R=Set^.Current) THEN Curr := HelpSet^.Current END;
        InsertUptoN(HelpSet,R^.DataPtr)
      END;
      (* delete the node in the source set and step forward *)
      R1 := R;
      R := R^.Thread;
      DEALLOCATE(R1^.DataPtr,Classes[Set^.ClassID].Bytes);
      DEALLOCATE(R1,TSIZE(ElementType))
    END;
    (* switch from source to target set *)
    Set := HelpSet;
    (* Restore the current element *)
    Set^.Current := Curr
  END SETCompact;

  PROCEDURE SETIn(    Set : SETType;
                  VAR Data: ARRAY OF BYTE): BOOLEAN;

  BEGIN
    RETURN(In(Set,ADR(Data)))
  END SETIn;

  PROCEDURE SETCard(    Set: SETType): CARDINAL;

  BEGIN
    RETURN(Set^.N)
  END SETCard;

  PROCEDURE SETCardUncompacted(    Set: SETType): CARDINAL;

  BEGIN
    (* return the number of physically allocated elements *)
    RETURN(Set^.N+Set^.DeletedN)
  END SETCardUncompacted;

  PROCEDURE SETCopy(    Source: SETType;
                    VAR Target: SETType);

  VAR
    R: ElementTypePtr;

  BEGIN
    (* create an optimal structure for the elements which are to be copied *)
    SETCreateForN(Target,Source^.ClassID,Source^.N);
    (* loop to insert the non-deleted elements into the target set *)
    R := Source^.ListThread;
    WHILE (R#NIL) DO
      IF (NOT R^.Deleted) THEN InsertUptoN(Target,R^.DataPtr) END;
      R := R^.Thread
    END
  END SETCopy;

  PROCEDURE SETIsSubset(    FirstSet : SETType;
                            SecondSet: SETType): BOOLEAN;

  VAR
    R: ElementTypePtr;

  BEGIN
    (* exit if the cardinality check is not ok. FirstSet can only be a subset
       of SecondSet if it does not have more elements *)
    IF (FirstSet^.N>SecondSet^.N) THEN RETURN(FALSE) END;
    (* loop until there is one element of FirstSet which is not contained
       in SecondSet *)
    R := FirstSet^.ListThread;
    WHILE (R#NIL) DO
      IF (NOT R^.Deleted) THEN
        IF (NOT In(SecondSet,R^.DataPtr)) THEN RETURN(FALSE) END
      END;
      R := R^.Thread
    END;
    RETURN(TRUE)
  END SETIsSubset;

  PROCEDURE SETIsStrictSubset(    FirstSet : SETType;
                                  SecondSet: SETType): BOOLEAN;

  BEGIN
    (* FirstSet can only be a strict subset of SecondSet if it does contain
       less elements *)
    IF (FirstSet^.N>=SecondSet^.N) THEN RETURN(FALSE) END;
    RETURN(SETIsSubset(FirstSet,SecondSet))
  END SETIsStrictSubset;

  PROCEDURE SETIsDisjunct(    FirstSet : SETType;
                              SecondSet: SETType): BOOLEAN;

  VAR
    R: ElementTypePtr;

  BEGIN
    (* loop until one element of FirstSet is in SecondSet *)
    R := FirstSet^.ListThread;
    WHILE (R#NIL) DO
      IF (NOT R^.Deleted) THEN
        IF (In(SecondSet,R^.DataPtr)) THEN RETURN(FALSE) END
      END;
      R := R^.Thread
    END;
    RETURN(TRUE)
  END SETIsDisjunct;

  PROCEDURE SETCompare(    FirstSet : SETType;
                           SecondSet: SETType;
                       VAR Equal    : BOOLEAN;
                       VAR Less     : BOOLEAN;
                       VAR Greater  : BOOLEAN);

  BEGIN
    Equal := FALSE;
    Less := FALSE;
    Greater := FALSE;
    (* first check for set cardinalities *)
    IF (FirstSet^.N<SecondSet^.N) THEN
      Less := SETIsSubset(FirstSet,SecondSet)
    ELSIF (FirstSet^.N>SecondSet^.N) THEN
      Greater := SETIsSubset(SecondSet,FirstSet)
    ELSE
      Equal := SETIsSubset(FirstSet,SecondSet)
          (* = SETIsSubset(SecondSet,FirstSet) ! *)
    END
  END SETCompare;

  PROCEDURE SETUnion(VAR FirstSet : SETType;
                         SecondSet: SETType);

  VAR
    R: ElementTypePtr;

  BEGIN
    (* insert all elements of the first into the second set *)
    R := SecondSet^.ListThread;
    WHILE (R#NIL) DO
      IF (NOT R^.Deleted) THEN Insert(FirstSet,R^.DataPtr) END;
      R := R^.Thread
    END
  END SETUnion;

  PROCEDURE SETIntersection(VAR FirstSet : SETType;
                                SecondSet: SETType);

  VAR
    R1,R2,P: ElementTypePtr;
    D      : CARDINAL;
    Result : SETCompareResultType;

  BEGIN
    (* loop over the first set *)
    R1 := FirstSet^.ListThread;
    WHILE (R1#NIL) DO
      (* always begin search at the root *)
      R2 := SecondSet^.TreeHead;
      IF (NOT R1^.Deleted) THEN
        (* search for data in the second set *)
        FindData(R2,R1^.DataPtr,Classes[FirstSet^.ClassID].Cmp,P,D,Result);
        IF (Result=SETEqual) THEN
          (* if the element is deleted, it will also be deleted in the
             first set *)
          IF (R2^.Deleted) THEN
            R1^.Deleted := TRUE;
            DEC(FirstSet^.N);
            INC(FirstSet^.DeletedN)
          END
        ELSE
          (* if the element is not in the second set, it will be deleted
             in the first *)
          R1^.Deleted := TRUE;
          DEC(FirstSet^.N);
          INC(FirstSet^.DeletedN)
        END
      END;
      R1 := R1^.Thread
    END
  END SETIntersection;

  PROCEDURE SETSymmetricDifference(VAR FirstSet : SETType;
                                       SecondSet: SETType);

  VAR
    Found  : BOOLEAN;
    Root   : ElementTypePtr;
    TempSet: SETType;

  BEGIN
    (* create a temporary help set *)
    SETCreate(TempSet,FirstSet^.ClassID);
    (* loop over SecondSet. After the loop FirstSet will contain only
       those elements which are not in SecondSet. TempSet will contain
       all those element which were deleted in FirstSet *)
    Root := SecondSet^.ListThread;
    WHILE (Root#NIL) DO
      IF (NOT Root^.Deleted) THEN
        DeleteElement(FirstSet,Root^.DataPtr,Found);
        IF (Found) THEN Insert(TempSet,Root^.DataPtr) END
      END;
      Root := Root^.Thread
    END;
    (* another loop over SecondSet. Now those elements of SecondSet are
       inserted into FirstSet which were not deleted in the first loop
       (that are all elements which are ONLY on SecondSet, NOT in FirstSet *)
    Root := SecondSet^.ListThread;
    WHILE (Root#NIL) DO
      IF (NOT Root^.Deleted) THEN
        IF (NOT In(TempSet,Root^.DataPtr)) THEN
          Insert(FirstSet,Root^.DataPtr)
        END
      END;
      Root := Root^.Thread
    END;
    (* release temporary set *)
    SETKill(TempSet)
  END SETSymmetricDifference;

  PROCEDURE SETDifference(VAR FirstSet : SETType;
                              SecondSet: SETType);

  VAR
    Root : ElementTypePtr;
    Found: BOOLEAN;

  BEGIN
    (* remove all SecondSet elements from FirstSet *)
    Root := SecondSet^.ListThread;
    WHILE (Root#NIL) DO
      IF (NOT Root^.Deleted) THEN
        DeleteElement(FirstSet,Root^.DataPtr,Found)
      END;
      Root := Root^.Thread
    END
  END SETDifference;

  PROCEDURE SETLoop(VAR Set: SETType);

  BEGIN
    (* a loop over the entire set requires a compare procedure which classifies
       all elements as SETEqual *)
    Set^.RangeCmp := FullLoopCmpProc;
    (* move element pointer to the start *)
    Set^.Current := Set^.ListThread
  END SETLoop;

  PROCEDURE SETGet(VAR Set  : SETType;
                   VAR Data : ARRAY OF BYTE;
                   VAR Found: BOOLEAN);

  VAR
    ElemPtr: ElementTypePtr;

  BEGIN
    (* Compute next element *)
    Get(Set,ElemPtr,Found);
    (* copy bytes from the set element to the output value, if an
       element could be found *)
    IF (Found) THEN
      Copy(ElemPtr^.DataPtr,ADR(Data),Classes[Set^.ClassID].Bytes)
    END
  END SETGet;

  PROCEDURE SETGotoElement(VAR Set  : SETType;
                           VAR Data : ARRAY OF BYTE;
                           VAR Found: BOOLEAN);

  VAR
    Root,P: ElementTypePtr;
    D     : CARDINAL;
    Result: SETCompareResultType;

  BEGIN
    (* locate data *)
    Root := Set^.TreeHead;
    FindData(Root,ADR(Data),Classes[Set^.ClassID].Cmp,P,D,Result);
    Found := (Result=SETEqual) AND (NOT Root^.Deleted);
    (* change pointer if found (and not deleted) *)
    IF (Found) THEN Set^.Current := Root END
  END SETGotoElement;

  PROCEDURE SETGotoMinimum(VAR Set: SETType);

  BEGIN
    Set^.Current := Set^.ListThread
  END SETGotoMinimum;

  PROCEDURE SETGotoMaximum(VAR Set: SETType);

  VAR
    Ptr: ElementTypePtr;

  BEGIN

    (* the largest set element can be found at the lower right leaf of the
       binary search tree *)
    Ptr := Set^.TreeHead;
    Set^.Current := NIL;
    WHILE (Ptr<>NIL) DO
      (* Store predecessor and go on *)
      Set^.Current := Ptr;
      Ptr := Ptr^.Right
    END
  END SETGotoMaximum;

  PROCEDURE SETRangeLoop(VAR Set        : SETType;
                             LoopCmpProc: SETCompareProcType;
                         VAR CmpData    : ARRAY OF BYTE);

  VAR
    Found: BOOLEAN;
    Next : ElementTypePtr;

  BEGIN
    (* store procedure address with the set *)
    Set^.RangeCmp := LoopCmpProc;
    (* find the first element which is classified as SETEqual. The first
       element to be returned can then be found in the left subtree under
       the computed one (if it is not already found) *)
    Set^.Current := Set^.TreeHead;
    Found := FALSE;
    (* loop until data is located or until end is reached *)
    WHILE (Set^.Current<>NIL) AND (NOT Found) DO
      CASE LoopCmpProc(ADR(CmpData),Set^.Current^.DataPtr) OF
        SETEqual  : Found := TRUE;
      | SETLess   : Set^.Current := Set^.Current^.Left;
      | SETGreater: Set^.Current := Set^.Current^.Right
      END
    END;
    (* exit if there is no SETEqual element *)
    IF (NOT Found) THEN RETURN END;
      (* Find the smallest SETEqual-element in the left partial tree under
	 the current element. Go down to the left, if an SETEqual-element
	 is found, go down to the right, if there is a SETGreater-element *)
    Next := Set^.Current^.Left;
    WHILE (Next#NIL) DO
      CASE LoopCmpProc(ADR(CmpData),Next^.DataPtr) OF
	(* Store the element as the smallest one (for the moment) *)
        SETEqual  : Set^.Current := Next; Next := Next^.Left;
	(* Go down to the right without updating the smallest element *)
      | SETGreater: Next := Next^.Right;
	(* Error if there is an illegal compare procedure *)
      | SETLess   :
		    HALT
      END
    END
  END SETRangeLoop;

  PROCEDURE SETRangeExtractAnyElement(VAR Set        : SETType;
                             	    	  LoopCmpProc: SETCompareProcType;
                         	      VAR CmpData    : ARRAY OF BYTE;
				      VAR Data	     : ARRAY OF BYTE;
				      VAR Found      : BOOLEAN);

  VAR
    ElemPtr: ElementTypePtr;

  BEGIN
    IF (Set^.N=0) THEN Found := FALSE; RETURN END;
    (* first try to find a non-deleted element without starting at the
       beginning --> efficiency+ *)
    Set^.RangeCmp := LoopCmpProc;
    Get(Set,ElemPtr,Found);
    (* start again if there's no element *)
    IF (NOT Found) THEN
      SETRangeLoop(Set,LoopCmpProc,CmpData);
      (* find first non-deleted element *)
      Get(Set,ElemPtr,Found)
    END;
    (* Copy & Delete if found. Perfom no automatic compactification because
       this would decrease the efficiency of SETExtractAnyElement too much *)
    IF (Found) THEN
      Copy(ElemPtr^.DataPtr,ADR(Data),Classes[Set^.ClassID].Bytes);
      DEC(Set^.N); INC(Set^.DeletedN); ElemPtr^.Deleted := TRUE
    END
  END SETRangeExtractAnyElement;
  
  PROCEDURE SETCheckConsistency(    Set: SETType);

  VAR
    CheckN       : CARDINAL;
    CheckDeletedN: CARDINAL;
    CheckDepth   : CARDINAL;
    Ptr          : ElementTypePtr;
    Prev         : ElementTypePtr;

    PROCEDURE InOrder(    Ptr     : ElementTypePtr;
			  D       : CARDINAL;
                      VAR N       : CARDINAL;
                      VAR DeletedN: CARDINAL);

    BEGIN
      (* exit if leaf is reached. Actualize maximal depth if necessary *)
      IF (Ptr=NIL) THEN
        IF (D>CheckDepth) THEN CheckDepth := D END;
        RETURN
      END;
      (* exit if one of the counters is expired *)
      IF (N>Set^.N) OR (DeletedN>Set^.DeletedN) THEN RETURN END;
      (* recursively go down to the left *)
      InOrder(Ptr^.Left,D+1,N,DeletedN);

      (* update counters *)
      IF (Ptr^.Deleted) THEN INC(DeletedN) ELSE INC(N) END;
      (* check for the left and right order *)
      IF (Ptr^.Left#NIL) THEN
        CASE Classes[Set^.ClassID].Cmp(Ptr^.DataPtr,Ptr^.Left^.DataPtr) OF
          SETGreater:
                        (* ok *)
        | SETEqual  :
            HALT;
        | SETLess   :
            HALT;
        END;
        CASE Classes[Set^.ClassID].Cmp(Ptr^.Left^.DataPtr,Ptr^.DataPtr) OF
          SETLess   :
                        (* ok *)
        | SETEqual  :
            HALT;
        | SETGreater:
            HALT;
        END;
      END;
      IF (Ptr^.Right#NIL) THEN
        CASE Classes[Set^.ClassID].Cmp(Ptr^.DataPtr,Ptr^.Right^.DataPtr) OF
          SETLess   :
                        (* ok *)
        | SETEqual  :
            HALT;
        | SETGreater:
            HALT;
        END;
        CASE Classes[Set^.ClassID].Cmp(Ptr^.Right^.DataPtr,Ptr^.DataPtr) OF
          SETGreater:
                        (* ok *)
        | SETEqual  :
            HALT;
        | SETLess   :
            HALT;
        END
      END;

      (* go down to the right *)
      InOrder(Ptr^.Right,D+1,N,DeletedN)
    END InOrder;

  BEGIN
    IF (Set=NIL) THEN
      HALT
    END;
    (* preset check counters *)
    CheckN := 0;
    CheckDeletedN := 0;
    CheckDepth := 0;
    (* loop over the linear list *)
    Prev := NIL;
    Ptr := Set^.ListThread;
    WHILE (Ptr#NIL) AND (CheckN<=Set^.N) AND (CheckDeletedN<=Set^.DeletedN) DO
      (* check the sorting order *)
      IF (Prev#NIL) THEN
        CASE Classes[Set^.ClassID].Cmp(Prev^.DataPtr,Ptr^.DataPtr) OF
          SETLess   : 
                        (* ok *)
        | SETEqual  :
            HALT
        | SETGreater:
            HALT
        END;
        CASE Classes[Set^.ClassID].Cmp(Ptr^.DataPtr,Prev^.DataPtr) OF
          SETGreater: 
                        (* ok *)
        | SETEqual  :
            HALT
        | SETLess   :
            HALT
        END
      END;
      (* update counters *)
      IF (Ptr^.Deleted) THEN INC(CheckDeletedN) ELSE INC(CheckN) END;
      (* goto next node *)
      Prev := Ptr;
      Ptr := Ptr^.Thread
    END;
    (* check counters *)
    IF (CheckN#Set^.N) THEN
      HALT
    END;
    IF (CheckDeletedN#Set^.DeletedN) THEN
      HALT
    END;
    (* reset counters for the second test *)
    CheckN := 0;
    CheckDeletedN := 0;
    (* loop over the tree in inorder way *)
    InOrder(Set^.TreeHead,0,CheckN,CheckDeletedN);
    (* check counters *)
    IF (CheckN#Set^.N) THEN
      HALT
    END;
    IF (CheckDeletedN#Set^.DeletedN) THEN
      HALT
    END;
    IF (CheckDepth#Set^.Depth) THEN
      HALT
    END;
  END SETCheckConsistency;

BEGIN ClassN := 0
END SETs.
